#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-09-20
# @Filename: corrections.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from lvmguider.actor import lvmguider_parser


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["corrections"]


@lvmguider_parser.command()
@click.argument("MODE", type=click.Choice(["enable", "disable"]))
async def corrections(command: GuiderCommand, mode: str):
    """Enables/disables corrections during guiding."""

    if not command.actor.guider:
        return command.fail("Guider is not running.")

    if mode.lower() == "enable":
        command.actor.guider.apply_corrections = True
    elif mode.lower() == "disable":
        command.actor.guider.apply_corrections = False
    else:
        return command.fail(f"Invalid mode {mode!r}.")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: guide.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

import click
import numpy

from lvmguider.actor import lvmguider_parser
from lvmguider.guider import CriticalGuiderError, Guider
from lvmguider.maskbits import GuiderStatus
from lvmguider.tools import wait_until_cameras_are_idle


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["guide"]


def is_stopping(command: GuiderCommand):
    return command.actor.status & GuiderStatus.STOPPING


@lvmguider_parser.command()
@click.argument("RA", type=float)
@click.argument("DEC", type=float)
@click.argument("PA", type=float, default=0.0, required=False)
@click.option(
    "--exposure-time",
    "-t",
    type=float,
    default=5,
    help="Exposure time.",
)
@click.option(
    "--reference-pixel",
    type=click.Tuple((float, float)),
    help="The pixel of the full frame to use as pointing reference.",
)
@click.option(
    "--guide-tolerance",
    type=float,
    help="The acquisition tolerance, in arcsec, after which guiding will be started.",
)
@click.option(
    "--apply-corrections/--no-apply-corrections",
    is_flag=True,
    default=True,
    show_default=True,
    help="Whether apply the measured corrections.",
)
@click.option(
    "--one",
    is_flag=True,
    help="Do one single iteration and exit.",
)
async def guide(
    command: GuiderCommand,
    ra: float,
    dec: float,
    pa: float = 0.0,
    exposure_time: float = 5.0,
    reference_pixel: tuple[float, float] | None = None,
    guide_tolerance: float | None = None,
    apply_corrections: bool = True,
    one: bool = False,
):
    """Starts the guide loop."""

    actor = command.actor
    MAX_EXPTIME: float = 18

    guider = Guider(
        command,
        (ra, dec, pa),
        pixel=reference_pixel,
        apply_corrections=apply_corrections,
    )
    command.actor.guider = guider

    if actor.status & GuiderStatus.NON_IDLE:
        return command.finish("Guider is not idle. Stop the guide loop.")

    await wait_until_cameras_are_idle(command)

    # Force the cameras to check the last image.
    command.actor.cameras.reset_seqno()

    while True:
        try:
            actor.guide_task = asyncio.create_task(
                guider.guide_one(
                    exposure_time,
                    guide_tolerance=guide_tolerance,
                )
            )
            await actor.guide_task
        except CriticalGuiderError as err:
            command.actor.status |= GuiderStatus.FAILED
            return command.fail(f"Stopping the guide loop due to critical error: {err}")
        except asyncio.CancelledError:
            # This means that the stop command was issued. All good.
            break
        except Exception as err:
            command.actor.status |= GuiderStatus.FAILED
            command.warning(f"Failed guiding with error: {err}")
            if "No solutions found" in str(err) and exposure_time < MAX_EXPTIME:
                exposure_time = numpy.clip(exposure_time * 1.5, 1, MAX_EXPTIME)
                exposure_time = numpy.round(exposure_time, 1)
                command.warning(f"Exposure time increased to {exposure_time:.1f} s")
        finally:
            if is_stopping(command):
                break

        if one:
            break

    actor.status = GuiderStatus.IDLE
    command.actor.guider = None

    return command.finish("The guide loop has finished.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-09-03
# @Filename: plotting.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib

from typing import TYPE_CHECKING, Callable, Sequence

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.text
import matplotlib.transforms as mtransforms
import numpy
import seaborn
from matplotlib.axes import Axes
from matplotlib.transforms import Bbox

from lvmguider import config


if TYPE_CHECKING:
    import pandas
    from matplotlib.figure import Figure

    from lvmguider.dataclasses import CoAdd_CameraSolution, GlobalSolution


matplotlib.use("agg")


AnyPath = str | os.PathLike


def plot_qa(
    solution: GlobalSolution,
    root: AnyPath | None = None,
    save_subplots: bool = True,
):
    """Produces guider QA plots from a global solution object."""

    root = pathlib.Path(root or (solution.path.parent / "qa"))
    root.mkdir(parents=True, exist_ok=True)

    stem = solution.path.stem

    # Set up seaborn and matplotlib.
    seaborn.set_theme(
        style="darkgrid",
        palette="deep",
        font="serif",
        font_scale=1.2,  # type: ignore
    )

    with plt.ioff():
        # Plot PA
        outpath_pa = root / (stem + "_pa.pdf")
        plot_position_angle(outpath_pa, solution, save_subplots=save_subplots)

        # Plot ZP
        outpath_zp = root / (stem + "_zp.pdf")
        plot_zero_point_or_fwhm(
            outpath_zp,
            solution,
            save_subplots=save_subplots,
            column="zero_point",
        )

        # Plot FWHM
        outpath_fwhm = root / (stem + "_fwhm.pdf")
        plot_zero_point_or_fwhm(
            outpath_fwhm,
            solution,
            save_subplots=save_subplots,
            column="fwhm",
        )

        # Plot guider offsets
        outpath_fwhm = root / (stem + "_guide_offsets.pdf")
        plot_guider_offsets(
            outpath_fwhm,
            solution,
            save_subplots=save_subplots,
        )

    seaborn.reset_orig()


def get_subplot_path(
    orig_path: AnyPath,
    suffix: str = "",
    extension="png",
    subdir: str = "subplots",
):
    """Creates a path for a subplot."""

    orig_path = pathlib.Path(orig_path)

    orig_parent = orig_path.parent
    stem = orig_path.stem

    path = orig_parent

    if subdir:
        path /= subdir

    return path / (stem + suffix + f".{extension}")


def create_subplot(func: Callable, path: AnyPath, *args, dpi: int = 100, **kwargs):
    """Creates a subplot by calling a plotting function."""

    pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)

    fig, ax = plt.subplots(figsize=(11, 8))
    func(ax, *args, **kwargs)
    fig.savefig(str(path), dpi=dpi)


def save_subplot(
    fig: Figure,
    axes: Axes | list[Axes],
    path: AnyPath,
    pad: float | Sequence[float] = (0.01, 0.0),
    bbox_range: Sequence[float] | None = None,
):
    """Saves a subplot, including axes labels, tick labels, and titles."""

    # https://stackoverflow.com/questions/4325733/save-a-subplot-in-matplotlib

    if isinstance(axes, Axes):
        axes = [axes]

    pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)

    if bbox_range is not None:
        fig.savefig(
            str(path),
            bbox_inches=mtransforms.Bbox(
                # This is in "figure fraction" for the bottom half
                # input in [[xmin, ymin], [xmax, ymax]]
                numpy.array(
                    [
                        [bbox_range[0], bbox_range[1]],
                        [bbox_range[2], bbox_range[3]],
                    ]
                )
            ).transformed(
                (fig.transFigure - fig.dpi_scale_trans)  # type: ignore
            ),
        )
        return

    items = []
    # For text objects, we need to draw the figure first, otherwise
    # the extents are undefined.
    for ax in axes:
        ax.figure.canvas.draw()

        items += ax.get_xticklabels()
        items += ax.get_yticklabels()
        items += [ax.get_xaxis().get_label(), ax.get_yaxis().get_label()]
        items += [ax, ax.title]  # type: ignore

    bbox = Bbox.union([item.get_window_extent() for item in items])

    if isinstance(pad, Sequence):
        bbox = bbox.expanded(1.0 + pad[0], 1.0 + pad[1])
    else:
        bbox = bbox.expanded(1.0 + pad, 1.0 + pad)

    extent = bbox.transformed(fig.dpi_scale_trans.inverted())  # type: ignore

    fig.savefig(str(path), bbox_inches=extent)


def get_camera_figure(
    left: float = 0.1,
    right: float = 0.9,
    bottom: float = 0.075,
    top: float = 0.91,
    wspace: float = 0.1,
    hspace: float = 0.2,
):
    """Returns a customised figure and axes."""

    fig, axd = plt.subplot_mosaic(
        [["east", "west"], ["global", "global"]],
        figsize=(11, 8),
    )
    fig.subplots_adjust(
        left=left,
        right=right,
        bottom=bottom,
        top=top,
        wspace=wspace,
        hspace=hspace,
    )

    for name, ax in axd.items():
        ax.axis("off")

        ax.set_xlabel("Frame number")

        if name == "global":
            ax.set_title("Full frame")
        else:
            ax.set_title(f"Camera {str(name).capitalize()}")

        ax.ticklabel_format(useOffset=False)
        ax.xaxis.get_major_locator().set_params(integer=True)

    return fig, axd


def plot_position_angle(
    outpath: pathlib.Path,
    solution: GlobalSolution,
    save_subplots: bool = False,
):
    """Plots the position angle."""

    fig, axd = get_camera_figure(wspace=0.4)

    # Plot top panels with the East and West camera PA and PA error.
    for camera_solution in solution.coadd_solutions:
        if camera_solution.frames is None or len(camera_solution.frames) == 0:
            continue

        camera = camera_solution.camera

        frame_data = camera_solution.frame_data()
        reprocessed = camera_solution.frames[0].reprocessed

        if not reprocessed:
            frame_data = frame_data.loc[frame_data.wcs_mode == "gaia"]

        if len(frame_data.pa.dropna()) < 2:
            continue

        # Unhide plot.
        ax = axd[camera]  # type: ignore
        ax.axis("on")

        # Plot PA.
        axes = _plot_pa_axes(
            ax,
            frame_data,
            pa_error_mode="mean",
            legend=False,
        )

        # Remove the inside labels since they are identical left and right.
        if camera == "east":
            axes[1].set_ylabel("")
        else:
            axes[0].set_ylabel("")

        ax.set_xlabel("")

        if save_subplots:
            create_subplot(
                _plot_pa_axes,
                get_subplot_path(outpath, f"_{camera}"),
                frame_data,
                title=f"Camera {camera.capitalize()}",
                pa_error_mode="mean",
                legend=True,
            )

    # Now use the PAs from the guider data (i.e., full frame).
    guider_data = solution.guider_data()
    guider_data = guider_data.loc[guider_data.guide_mode == "guide"]

    if len(guider_data.pa.dropna()) >= 2:
        ax = axd["global"]  # type: ignore
        ax.axis("on")

        _plot_pa_axes(ax, guider_data, pa_error_mode="mean", legend=True)

        if save_subplots:
            create_subplot(
                _plot_pa_axes,
                get_subplot_path(outpath),
                guider_data,
                title="Full frame",
                pa_error_mode="mean",
                legend=True,
            )

    fig.suptitle(f"Position angle for {solution.path.name}")
    fig.savefig(str(outpath))

    plt.close("all")


def _plot_pa_axes(
    ax: Axes,
    data: pandas.DataFrame,
    title: str = "",
    pa_error_mode: str = "first",
    legend: bool = True,
):
    """Plot PA axes."""

    ax.ticklabel_format(useOffset=True)

    pa_threshold = config["coadds"]["warnings"]["pa_error"]

    # Create right axis for the PA error.
    right_ax = ax.twinx()
    right_ax.ticklabel_format(useOffset=True)

    # Plot absolute position angle.
    (pa_plot,) = ax.plot(
        data.frameno,
        data.pa,
        "b-",
        zorder=20,
        label="Position angle",
    )

    # For the cameras we calculate the error wrt the mean PA.
    # For the full frame PAs we use the initial PA and calculate the "drift".
    if pa_error_mode == "first":
        ref_pa = data.pa.dropna().iloc[0]
    else:
        ref_pa = data.pa.dropna().mean()

    # Plot PA error.
    (pa_error_plot,) = right_ax.plot(
        data.frameno,
        ref_pa - data.pa,
        "g-",
        zorder=20,
        label="PA error",
    )

    # If the PA error goes over the limit, show a band with the threshold.
    abs_error = (ref_pa - data.pa).abs().max()
    if abs_error > pa_threshold:
        right_ax.axhspan(
            ymin=-pa_threshold,
            ymax=pa_threshold,
            facecolor="red",
            edgecolor="None",
            linestyle="None",
            alpha=0.2,
            zorder=5,
        )

    if legend:
        ax.legend(
            handles=[pa_plot, pa_error_plot],
            loc="upper left",
        ).set_zorder(100)
        ax.margins(0.05, 0.2)

    # The left and right y-axis are different z-stacks so it's not possible to mix
    # and match. This moves the right y-axis to the background.
    right_ax.set_zorder(-1)
    ax.patch.set_visible(False)  # type: ignore
    ax.grid(False)
    right_ax.patch.set_visible(True)  # type: ignore

    ax.set_xlabel("Frame number")
    ax.set_ylabel("Position angle [deg]", labelpad=10)
    right_ax.set_ylabel("Position angle error [deg]", labelpad=10)

    if title:
        ax.set_title(title)

    return [ax, right_ax]


def plot_zero_point_or_fwhm(
    outpath: pathlib.Path,
    solution: GlobalSolution,
    save_subplots: bool = False,
    column="zero_point",
):
    """Plots the zero point or FWHM of an exposure."""

    fig, axd = get_camera_figure()

    # Get guider data first. We'll use this to select camera frames to use.
    guider_data = solution.guider_data()
    guider_data = guider_data.loc[guider_data.guide_mode == "guide"]

    # Plot top panels with the East and West camera zero point or FWHM.
    for camera_solution in solution.coadd_solutions:
        camera = camera_solution.camera

        # Get all the frames and the ZP/FWHM value of the co-added camera image.
        frame_data = camera_solution.frame_data()

        # We want to reject the frames in which we were acquiring since those may
        # have starts blurred or moving. If there's at least one "gaia" in the
        # wcs_mode, use that. Otherwise this must be a reprocessed set of frames
        # which all have "astrometry.net". Then use the guider_data framenos.
        if "gaia" in frame_data.wcs_mode.values:
            frame_data = frame_data.loc[frame_data.wcs_mode == "gaia", :]
        else:
            valid = numpy.in1d(frame_data.frameno, guider_data.frameno)
            frame_data = frame_data.loc[valid]

        coadd_value = getattr(camera_solution, column)

        if len(frame_data) == 0 and numpy.isnan(coadd_value):
            continue

        # Un-hide the axes.
        ax = axd[camera]  # type: ignore
        ax.axis("on")

        # Plot data.
        ax = _plot_zero_point_or_fwhm_axes(
            ax,
            camera_solution,
            frame_data,
            coadd_value,
            column=column,
            legend=False,
        )

        # Move the West camera y-axis to the right to unclutter the central section.
        if camera == "west":
            ax.yaxis.tick_right()
            ax.yaxis.set_label_position("right")

        ax.set_xlabel("")

        if save_subplots:
            create_subplot(
                _plot_zero_point_or_fwhm_axes,
                get_subplot_path(outpath, f"_{camera}"),
                camera_solution,
                frame_data,
                coadd_value,
                column=column,
                title=f"Camera {camera.capitalize()}",
                legend=True,
            )

    # Now do the full frame. Almost identical to above.
    global_value = getattr(solution, column)

    if len(guider_data.zero_point.dropna()) >= 2 or not numpy.isnan(global_value):
        ax = axd["global"]  # type: ignore
        ax.axis("on")

        _plot_zero_point_or_fwhm_axes(
            ax,
            solution,
            guider_data,
            global_value,
            column=column,
            legend=True,
        )

        if save_subplots:
            create_subplot(
                _plot_zero_point_or_fwhm_axes,
                get_subplot_path(outpath),
                solution,
                guider_data,
                global_value,
                column=column,
                title="Full frame",
                legend=True,
            )

    if column == "zero_point":
        fig.suptitle(f"Zero point for {solution.path.name}")
    else:
        fig.suptitle(f"FWHM for {solution.path.name}")

    fig.savefig(str(outpath))

    plt.close("all")


def _plot_zero_point_or_fwhm_axes(
    ax: Axes,
    solution: GlobalSolution | CoAdd_CameraSolution,
    data: pandas.DataFrame,
    coadd: float,
    column="zero_point",
    title: str | None = None,
    legend: bool = True,
):
    """Plot zero point or FWHM axes."""

    handles = []

    # Plot the data for each frame.
    (plot,) = ax.plot(
        data.frameno,
        data[column],
        "b-",
        label="Frames",
    )
    handles.append(plot)

    # If the co-added value exists, plot it as a horizontal dashed line.
    if not numpy.isnan(coadd):
        median_line = ax.axhline(
            y=coadd,
            color="k",
            linestyle="dashed",
            label="Co-add",
            alpha=0.5,
        )
        handles.append(median_line)

    # Add label with warnings.
    if len(data) > 0 and data["telescope"].iloc[0] != "spec":
        if column == "zero_point":
            warn = solution.transp_warning()
        else:
            warn = solution.fwhm_warning()
    else:
        warn = False

    if warn:
        ax.text(
            0.98,
            0.94,
            "WARNING",
            transform=ax.transAxes,
            fontsize=14,
            ha="right",
            va="center",
            color="r",
        )

    if legend:
        loc = "lower right" if column == "zero_point" else "upper left"
        ax.legend(handles=handles, loc=loc).set_zorder(100)
        ax.margins(0.05, 0.25)

    ax.set_xlabel("Frame number")

    if column == "zero_point":
        ax.set_ylabel("Zero Point [mag]", labelpad=10)
    else:
        ax.set_ylabel("FWHM [arcsec]", labelpad=10)

    if title:
        ax.set_title(title)

    return ax


def plot_guider_offsets(
    outpath: pathlib.Path,
    solution: GlobalSolution,
    save_subplots: bool = False,
):
    """Plots guider data (RA/Dec/PA offsets and applied corrections)."""

    fig, axd = plt.subplot_mosaic(
        [["sep", "sep"], ["meas", "applied"]],
        figsize=(11, 8),
    )

    fig.subplots_adjust(
        left=0.1,
        right=0.9,
        bottom=0.075,
        top=0.93,
        wspace=0.1,
        hspace=0.3,
    )

    gdata = solution.guider_data()
    gdata = gdata.loc[gdata.guide_mode == "guide"]

    # Top panel. On the left y axis plot the separation to the field
    # centre. In the right y axis plot the PA error (field - measured).
    gdata["separation"] = numpy.hypot(gdata.ra_off, gdata.dec_off)
    sep_data = gdata.loc[:, ["frameno", "separation"]]

    sep_ax = axd["sep"]  # type: ignore
    (sep_plot,) = sep_ax.plot(sep_data.frameno, sep_data.separation, "b-")
    sep_ax.set_xlabel("Frame number")
    sep_ax.set_ylabel("Separation [arcsec]", labelpad=10)

    sep_ax_r = sep_ax.twinx()
    pa_off = gdata.pa_field - (gdata.pa - 180)
    (pa_off_plot,) = sep_ax_r.plot(gdata.frameno, pa_off, "g-")
    sep_ax_r.set_ylabel("Position angle error [deg]", labelpad=10)
    sep_ax_r.legend(
        handles=[sep_plot, pa_off_plot],
        labels=["Separation", "PA error"],
        loc="upper right",
    )
    sep_ax_r.grid(False)

    # Bottom left panel. Plot the measured offsets in RA, Dec..
    meas_ax = axd["meas"]  # type: ignore
    meas_ax.plot(gdata.frameno, gdata.ra_off, "b-", label="RA", linewidth=0.5)
    meas_ax.plot(gdata.frameno, gdata.dec_off, "r-", label="Dec", linewidth=0.5)
    meas_ax.legend(loc="lower left")
    meas_ax.set_xlabel("Frame number")
    meas_ax.set_ylabel("Measured error [arcsec]", labelpad=10)

    # Bottom right panel. Plot the applied corrections in RA, Dec.
    # No point in plotting PA corrections since they are not applied
    # during guiding and we have excluded acquisition.
    appl_ax = axd["applied"]  # type: ignore
    appl_ax.plot(gdata.frameno, gdata.ax0_applied, "b-", linewidth=0.5)
    appl_ax.plot(gdata.frameno, gdata.ax1_applied, "r-", linewidth=0.5)
    appl_ax.set_xlabel("Frame number")
    appl_ax.set_ylabel("Applied offset [arcsec]")
    appl_ax.yaxis.tick_right()
    appl_ax.yaxis.set_label_position("right")

    if save_subplots:
        save_subplot(
            fig,
            [sep_ax, sep_ax_r],
            get_subplot_path(outpath, ""),
            bbox_range=[0.0, 0.47, 1.0, 0.98],
        )

        save_subplot(
            fig,
            [meas_ax],
            get_subplot_path(outpath, "_measured"),
            bbox_range=[0.0, 0.0, 0.505, 0.48],
        )

        save_subplot(
            fig,
            [appl_ax],
            get_subplot_path(outpath, "_applied"),
            bbox_range=[0.505, 0.0, 1.0, 0.48],
        )

    fig.suptitle(f"Guider data for {solution.path.name}")
    fig.savefig(str(outpath))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-10
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

from clu.actor import AMQPActor

from lvmguider import __version__, config, log
from lvmguider.cameras import Cameras
from lvmguider.maskbits import GuiderStatus


if TYPE_CHECKING:
    from lvmguider.guider import Guider


__all__ = ["LVMGuiderActor"]


class LVMGuiderActor(AMQPActor):
    """The ``lvmguider`` actor."""

    def __init__(self, *args, **kwargs):
        # The package imports a generic config file without actor or telescope info.
        # We expect the actor is initialised with it.
        aconfig = kwargs.get("config", None)
        if aconfig is None:
            raise RuntimeError("Actor must be initialised from a configuration file.")

        # Check that there's an actor section.
        if "actor" not in aconfig:
            raise RuntimeError("The configuration file does not have an actor section.")

        # Update package config.
        config._BASE = dict(config)
        config.update(aconfig)

        name = aconfig["actor"]["name"]
        self.telescope: str = aconfig.get("telescope", name.split(".")[1])

        # Update the config that will be set in the actor instance.
        kwargs["config"] = dict(config)

        super().__init__(*args, log=log, version=__version__, **kwargs)

        if self.model and self.model.schema:
            self.model.schema["additionalProperties"] = True

        self.cameras = Cameras(self.telescope)

        self._status = GuiderStatus.IDLE

        self.guider: Guider | None = None
        self.guide_task: asyncio.Task | None = None

        # Track model of the cameras and focuser associated to this telescope.
        self.models.actors.append(f"lvm.{self.telescope}.foc")

    @property
    def status(self):
        """Returns the guider status."""

        return self._status

    @status.setter
    def status(self, new_value: GuiderStatus):
        """Sets a new status and reports it to the users."""

        if new_value.value != self._status.value:
            self._status = new_value
            self.write(
                "d",
                {
                    "status": f"0x{self._status.value:x}",
                    "status_labels": ",".join(self._status.get_names()),
                },
                internal=True,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: transformations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
from datetime import datetime

from typing import Any, cast

import numpy
import pandas
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from astropy.utils.iers import conf
from astropy.wcs import WCS
from astropy.wcs.utils import fit_wcs_from_points
from scipy.spatial import KDTree

from lvmguider import config
from lvmguider.astrometrynet import AstrometrySolution, astrometrynet_quick
from lvmguider.extraction import extract_sources
from lvmguider.types import ARRAY_1D_F32, ARRAY_2D_F32


# Prevent astropy from downloading data.
conf.auto_download = False
conf.iers_degraded_accuracy = "ignore"


def ag_to_full_frame(
    camera: str,
    in_locs: numpy.ndarray,
    rot_shift: tuple | None = None,
    verbose=False,
):
    """Rotate and shift the star locations from an AG frame to the correct locations in
    the large "full frame" (master frame) corresponding to the full focal plane.

    Adapted from Tom Herbst's code.

    Parameters
    ----------
    camera:
        Which AG camera: ``skyw-e``, ``skyw-w``, ``sci-c``, ``sci-e``, ``sci-w``,
        ``skye-e``, ``skye-w``, or ``spec-e``.
    in_locs
        Numpy 2D array of (x, y) star locations.
    rot_shift
        List of input ``[theta_deg, Sx, Sz]` to use. If `None`, use best fit values

    Returns
    -------
    rs_loc
        Array of corresponding star locations in "full frame".

    """

    camera = camera.lower()

    # Get best-fit rotation and shift, based on NonLinLSQ_Fit.py
    if rot_shift is None:  # Use Best-Fit values from Suff_AGC.docx
        # Best-fit rotations and shifts
        # Last one is Sci on-axis for 21 Feb 23 - see Doc
        met_data = {
            "skyw-w": [-89.85, -76.53, 430.64],
            "skyw-e": [90.16, 3512.90, 433.74],
            "sci-w": [-89.84, -28.26, 415.33],
            "sci-e": [89.98, 3512.74, 447.25],
            "spec-e": [89.60, 3193.11, 504.26],
            "skye-w": [-90.69, -79.54, 436.38],
            "skye-e": [89.71, 3476.00, 457.13],
            "sci-c": [-89.995, 1707.4, 426.62],
        }

        # Pull out correct best-fit rotation and shift
        bF = met_data[camera]
        th, Sx, Sz = (numpy.radians(bF[0]), bF[1], bF[2])

    else:
        # Use supplied values
        th, Sx, Sz = (numpy.radians(rot_shift[0]), rot_shift[1], rot_shift[2])

    if verbose:
        print("")
        print("ag_to_full_frame: rot_shift for ", camera, numpy.degrees(th), Sx, Sz)
        print("")

    # Rotation matrix
    M = numpy.array(([numpy.cos(th), -numpy.sin(th)], [numpy.sin(th), numpy.cos(th)]))

    # 2xnPts rotation offset
    rOff = numpy.tile(numpy.array([[800.0, 550.0]]).transpose(), (1, in_locs.shape[0]))

    # Make 2xnPts Sx,Sz offset matrix
    off = numpy.tile(numpy.array([[Sx, Sz]]).transpose(), (1, in_locs.shape[0]))

    # Shift to center, Rotate, shift back, and offset
    rsLoc = numpy.dot(M, in_locs.T - rOff) + rOff + off

    # Return calculated positions (need to transpose for standard layout)
    return (rsLoc.T, (th, Sx, Sz))


def full_frame_to_ag(camera: str, locs: numpy.ndarray):
    """Transforms the star locations from the full frame
    to the internal pixel locations in the AG frame.

    Adapted from Tom Herbst.

    Parameters
    ----------
    camera:
        Which AG camera: ``skyw-e``, ``skyw-w``, ``sci-c``, ``sci-e``, ``sci-w``,
        ``skye-e``, ``skye-w``, or ``spec-e``.
    locs
        Numpy 2D array of (x, y) star locations.
    Returns
    -------
    ag_loc
        Array of corresponding star locations in the AG frame.

    """

    camera = camera.lower()

    met_data = {
        "skyw-w": [-89.85, -76.53, 430.64],
        "skyw-e": [90.16, 3512.90, 433.74],
        "sci-w": [-89.84, -28.26, 415.33],
        "sci-e": [89.98, 3512.74, 447.25],
        "spec-e": [89.60, 3193.11, 504.26],
        "skye-w": [-90.69, -79.54, 436.38],
        "skye-e": [89.71, 3476.00, 457.13],
        "sci-c": [-89.995, 1707.4, 426.62],
    }

    # Pull out correct best-fit rotation and shift
    bF = met_data[camera]
    th, Sx, Sz = (numpy.radians(bF[0]), bF[1], bF[2])

    # Rotation matrix
    M = numpy.array(([numpy.cos(th), -numpy.sin(th)], [numpy.sin(th), numpy.cos(th)]))

    # 2xnPts rotation offset
    rOff = numpy.tile(numpy.array([[800.0, 550.0]]).transpose(), (1, locs.shape[0]))

    # Make 2xnPts Sx,Sz offset matrix
    off = numpy.tile(numpy.array([[Sx, Sz]]).transpose(), (1, locs.shape[0]))

    # Transform FF --> AG. See Single_Sensor_Solve doc
    rsLoc = numpy.dot(M, (locs.T - rOff - off)) + rOff

    # Return calculated positions (need to transpose for standard layout)
    return rsLoc.T


def solve_locs(
    locs: pandas.DataFrame,
    ra: float,
    dec: float,
    full_frame=True,
    index_paths: dict[int, str] | None = None,
    scales: dict[int, list[int]] | None = None,
    output_root: str | None = None,
    verbose: bool = False,
    raise_on_unsolved: bool = True,
    **astrometrynet_kwargs,
):
    """Use astrometry.net to solve field, given a series of star locations.

    Adapted from Tom Herbst's code.

    Parameters
    ----------
    locs
        A Pandas DataFrame of ``x``, ``y``, and ``flux`` star locations.
    ra
        The estimated RA of the field.
    dec
        The estimated Dec of the field.
    full_frame
        Whether this is a full "full frame" field, requiring a different set of
        index files.
    index_paths
        The paths to the index files to use. A dictionary of series number
        to path on disk, e.g.,
        ``{5200: "/data/astrometrynet/5200", 4100: "/data/astrometrynet/5100"}``.
    scales
        Index scales to use. Otherwise uses all index files. The format
        is a dictionary of series number to a list of scales, e.g.,
        ``{5200: [4, 5], 4100: [10, 11]}``.
    output_root
        The path where to write the astrometry.net files. If `None`, uses temporary
        files.
    verbose
        Output additional information.
    raise_on_unsolved
        Whether to raise an error if the field was not solved.
    astrometrynet_kwargs
        Arguments to pass directly to `.astrometrynet_quick`.

    Returns
    -------
    myWCS
        A WCS entity (or `None` if it fails) with appropriate location information.
    solveInfo
        A list containing information about the solve:
        solveInfo[0] - [raSlv,deSlv]  RA,Dec of central pixel
                 [1] - angSlv         Rotation of field
                 [2] - pxScl          Arcsec / pixel of solve

    """

    lower_bound = 0.8  # Pixel scale hints (arcsec/pixel)
    upper_bound = 1.2
    radius = 5  # Search radius in degrees

    if full_frame:
        midX, midZ = config["xz_full_frame"]  # Middle of full Frame
        index_paths_default = {
            5200: "/data/astrometrynet/5200",
            4100: "/data/astrometrynet/4100",
        }
    else:
        midX, midZ = config["xz_ag_frame"]  # Middle of AG camera Frame
        index_paths_default = {5200: "/data/astrometrynet/5200"}

    index_paths = index_paths or index_paths_default
    scales = scales or {5200: [4, 5, 6]}

    locs = locs.copy()
    if full_frame:
        locs = locs.rename(columns={"x_full": "x", "y_full": "y"})

    solution = astrometrynet_quick(
        index_paths,
        locs,
        ra=ra,
        dec=dec,
        radius=radius,
        pixel_scale=(upper_bound + lower_bound) / 2,
        pixel_scale_factor_hi=upper_bound,
        pixel_scale_factor_lo=lower_bound,
        width=midX * 2,
        height=midZ * 2,
        scales=scales,
        raise_on_unsolved=raise_on_unsolved,
        output_root=output_root,
        **astrometrynet_kwargs,
    )

    if verbose:
        print("")
        print("Solution has match ? ", solution.solved)

    return solution


def radec2azel(raD, decD, lstD, site: EarthLocation | None = None):
    """Returns the Azimuth and Elevation for the supplied RA, Dec, and sidereal time.

    From Tom Herbst and Florian Briegel.

    Parameters
    ----------
    ra,dec
        Right Ascension and Declination in decimal degrees.
    lst
        Local Sidereal Time in decimal degrees.

    Returns
    -------
    az,el
        Azimuth and Elevation (Altitude) in decimal degrees

    """

    if site is None:
        site = EarthLocation.from_geodetic(
            lon=-70.70166667,
            lat=-29.00333333,
            height=2282.0,
        )

    assert isinstance(site, EarthLocation)

    lat_r = numpy.radians(site.lat.deg)

    ra, dec, lst = (
        numpy.radians(raD),
        numpy.radians(decD),
        numpy.radians(lstD),
    )  # Convert to radians

    ha = lst - ra

    el = numpy.arcsin(
        numpy.sin(dec) * numpy.sin(lat_r)
        + numpy.cos(dec) * numpy.cos(lat_r) * numpy.cos(ha)
    )

    rat = (numpy.sin(dec) - numpy.sin(el) * numpy.sin(lat_r)) / (
        numpy.cos(el) * numpy.cos(lat_r)
    )  # Ratio - need to pin [-1,1]

    if rat < -1.0:  # Goes wonky if roundoff puts it outside [1,1]
        rat = -1.0

    if rat > 1.0:
        rat = 1.0

    if numpy.sin(ha) < 0.0:
        az = numpy.arccos(rat)
    else:
        az = 2.0 * numpy.pi - numpy.arccos(rat)

    return numpy.degrees(az), numpy.degrees(el)


def azel2sazsel(azD, elD):
    """Returns the siderostat coordinates (saz, Sel) for the supplied Az-El.

    From Tom Herbst and Florian Briegel.

    Parameters
    ----------
    az,el
        Azimuth and Elevation (Altitude) in decimal degrees

    Returns
    -------
    sazD,selD
        Siderostat angles in degrees

    """

    r90 = numpy.radians(90.0)  # 90 deg in radians
    az, el = numpy.radians(azD), numpy.radians(elD)  # Convert to radians
    SEl = numpy.arccos(numpy.cos(el) * numpy.cos(az)) - r90  # SEl in radians
    rat = numpy.sin(el) / numpy.cos(SEl)  # Ratio
    if azD < 180.0:
        SAz = r90 - numpy.arcsin(rat)  # saz in radians
    else:
        SAz = numpy.arcsin(rat) - r90

    return numpy.degrees(SAz), numpy.degrees(SEl)  # Return values in degrees


def delta_radec2mot_axis(
    ra_ref,
    dec_ref,
    ra_new,
    dec_new,
    date: datetime | None = None,
    site: EarthLocation | None = None,
):
    """RA/Dec offset to motor axes.

    From Tom Herbst and Florian Briegel.

    """

    if site is None:
        site = EarthLocation.from_geodetic(
            lon=-70.70166667,
            lat=-29.00333333,
            height=2282.0,
        )

    assert isinstance(site, EarthLocation)

    date = date or datetime.utcnow()
    observing_time = Time(date, scale="utc", location=site)
    lst = observing_time.sidereal_time("mean")

    ref_az_d, ref_el_d = radec2azel(ra_ref, dec_ref, lst.deg, site=site)
    new_az_d, new_el_d = radec2azel(ra_new, dec_new, lst.deg, site=site)

    ref_saz_d, ref_sel_d = azel2sazsel(ref_az_d, ref_el_d)
    new_saz_d, new_sel_d = azel2sazsel(new_az_d, new_el_d)

    saz_diff_d = ref_saz_d - new_saz_d
    sel_diff_d = ref_sel_d - new_sel_d

    # To arcsec
    saz_diff_arcsec = saz_diff_d * -3600.0
    sel_diff_arcsec = sel_diff_d * -3600.0

    return saz_diff_arcsec, sel_diff_arcsec


def solve_camera_with_astrometrynet(
    input: pandas.DataFrame | pathlib.Path | str,
    ra: float,
    dec: float,
    solve_locs_kwargs: dict | None = None,
) -> AstrometrySolution:
    """Astrometrically solves a single camera using ``astrometry.net``.

    Parameters
    ----------
    input
        The input data. This is normally a Pandas data frame, the result of
        running `.extract_sources` on an image. It can also be a path to
        a file, in which case sources will be extracted here.
    ra,dec
        An initial guess of the RA and Dec of the centre of the camera.
    solve_locs_kwargs
        A dictionary of kwargs to pass to `.solve_locs`.

    Returns
    -------
    solution
       An object containing the astrometric solution.

    """

    solve_locs_kwargs = solve_locs_kwargs if solve_locs_kwargs is not None else {}

    if isinstance(input, (str, pathlib.Path)):
        input = extract_sources(input)

    sources = input.copy()

    locs_i = sources.loc[:, ["x", "y", "flux"]].copy()
    locs_i.loc[:, ["x", "y"]] -= 0.5  # We want centre of pixel to be (0.5, 0.5)

    camera_solution = solve_locs(
        locs_i,
        ra,
        dec,
        full_frame=False,
        raise_on_unsolved=False,
        **solve_locs_kwargs.copy(),
    )

    return camera_solution


def get_crota2(wcs: WCS):
    """Determines the ``CROTA2`` angle, in degrees, from a CD matrix."""

    cd: ARRAY_2D_F32

    if wcs.wcs.has_crota():
        return wcs.wcs.crota[1]
    elif wcs.wcs.has_cd():
        cd = wcs.wcs.cd
    elif wcs.wcs.has_pc():
        cd = wcs.wcs.pc * wcs.wcs.cdelt
    else:
        raise ValueError("WCS does not have information to determine CROTA2.")

    crota2 = numpy.degrees(numpy.arctan2(-cd[0, 1], cd[1, 1]))
    return float(crota2 if crota2 > 0 else crota2 + 360)


def match_with_gaia(
    wcs: WCS,
    sources: pandas.DataFrame,
    gaia_sources: pandas.DataFrame | None = None,
    max_separation: float = 2,
    concat: bool = False,
    db_connection_params: dict[str, Any] = {},
) -> tuple[pandas.DataFrame, int]:
    """Match detections to Gaia sources using nearest neighbours.

    Parameters
    ----------
    wcs
        The WCS associated with the detections. Used to determine pixels on the
        image frame.
    sources
        A data frame with extracted sources.
    gaia_sources
        A data frame with Gaia sources to be matched. If `None`, Gaia sources
        will be queried from the database.
    max_separation
        Maximum separation between detections and matched Gaia sources, in arcsec.
    concat
        If `True`, the returned data frame is the input ``sources`` concatenated
        with the match information.
    db_connection_params
        Database connection parameters to pass to `.get_gaia_sources` if
        ``gaia_sources`` are not passed.

    Returns
    -------
    matches
        A tuple with the matched data frame and the number of matches.

    """

    from lvmguider.tools import get_gaia_sources

    PIXSCALE = 1.009  # arcsec/pix

    # Epoch difference with Gaia DR3.
    GAIA_EPOCH = 2016.0
    epoch = Time.now().jyear
    epoch_diff = epoch - GAIA_EPOCH

    if gaia_sources is None:
        gaia_sources = get_gaia_sources(wcs, db_connection_params=db_connection_params)

    sources = sources.copy()
    gaia_sources = gaia_sources.copy()

    # Match detections with Gaia sources. Take into account proper motions.
    ra_gaia: ARRAY_1D_F32 = gaia_sources.ra.to_numpy(numpy.float32)
    dec_gaia: ARRAY_1D_F32 = gaia_sources.dec.to_numpy(numpy.float32)
    pmra: ARRAY_1D_F32 = gaia_sources.pmra.to_numpy(numpy.float32)
    pmdec: ARRAY_1D_F32 = gaia_sources.pmdec.to_numpy(numpy.float32)

    pmra_gaia = numpy.nan_to_num(pmra) / 1000 / 3600  # deg/yr
    pmdec_gaia = numpy.nan_to_num(pmdec) / 1000 / 3600

    ra_epoch = ra_gaia + pmra_gaia / numpy.cos(numpy.radians(dec_gaia)) * epoch_diff
    dec_epoch = dec_gaia + pmdec_gaia * epoch_diff

    gaia_sources["ra_epoch"] = ra_epoch
    gaia_sources["dec_epoch"] = dec_epoch

    # Calculate x/y pixels of the Gaia detections. We use origin 0 but the
    # sep/SExtractor x/y in sources assume that the centre of the lower left
    # pixel is (1,1) so we adjust the returned pixel values.
    xpix, ypix = wcs.wcs_world2pix(ra_epoch, dec_epoch, 0)
    gaia_sources["xpix"] = xpix + 0.5
    gaia_sources["ypix"] = ypix + 0.5

    tree = KDTree(gaia_sources.loc[:, ["xpix", "ypix"]].to_numpy())
    dd, ii = tree.query(sources.loc[:, ["x", "y"]].to_numpy())
    valid = dd < max_separation

    # Get Gaia rows for the valid matches. Change their indices to those
    # of their matching sources (which are 0..len(sources)-1 since we reindexed).
    matches: pandas.DataFrame = gaia_sources.iloc[ii[valid]].copy()
    assert isinstance(matches, pandas.DataFrame)

    matches.index = pandas.Index(numpy.arange(len(ii))[valid])

    dx = (matches.xpix - sources.x).astype(numpy.float32)
    dy = (matches.ypix - sources.y).astype(numpy.float32)
    match_sep = cast(pandas.Series, numpy.hypot(dx, dy) * PIXSCALE)
    matches.loc[:, "match_sep"] = match_sep.loc[matches.index]

    matches = matches.drop(columns=["xpix", "ypix"])

    if concat:
        sources.loc[:, matches.columns] = matches
        return sources, valid.sum()

    return matches, valid.sum()


def wcs_from_gaia(sources: pandas.DataFrame, xy_cols: list[str] = ["x", "y"]) -> WCS:
    """Creates a WCS from Gaia-matched sources."""

    # Get useful columns. Drop NaNs.
    matched_sources = sources.loc[:, ["ra_epoch", "dec_epoch", *xy_cols]]
    matched_sources.dropna(inplace=True)

    if len(matched_sources) < 5:
        raise RuntimeError("Insufficient number of Gaia matches.")

    skycoords = SkyCoord(
        ra=matched_sources.ra_epoch,
        dec=matched_sources.dec_epoch,
        unit="deg",
        frame="icrs",
    )

    return fit_wcs_from_points(
        (matched_sources[xy_cols[0]], matched_sources[xy_cols[1]]),
        skycoords,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: camera.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import pathlib
import re

from typing import TYPE_CHECKING

import numpy
import pandas
from astropy.io import fits

from sdsstools.time import get_sjd

from lvmguider import __version__
from lvmguider.extraction import extract_sources as extract_sources_func
from lvmguider.maskbits import GuiderStatus
from lvmguider.tools import (
    elapsed_time,
    header_from_model,
    run_in_executor,
    update_fits,
)


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["Cameras"]


class Cameras:
    """Exposures and solves the telescope cameras."""

    def __init__(self, telescope: str):
        self.telescope = telescope

        self.agcam = f"lvm.{telescope}.agcam"

        self.pwi = f"lvm.{telescope}.pwi"
        self.km = f"lvm.{telescope}.km"
        self.foc = f"lvm.{telescope}.foc"

        self.sjd = get_sjd("LCO")
        self.last_seqno: int = -1

        self.dark_file: dict[str, str] = {}

    async def expose(
        self,
        command: GuiderCommand,
        exposure_time: float = 5.0,
        flavour: str = "object",
        extract_sources: bool = False,
        nretries: int = 3,
    ) -> tuple[list[pathlib.Path], int, list[pandas.DataFrame] | None]:
        """Exposes the cameras and returns the filenames."""

        command.actor._status &= ~GuiderStatus.IDLE
        command.actor.status |= GuiderStatus.EXPOSING

        focus_position = await self._get_focus_position(command)

        next_seqno = self.get_next_seqno()

        command.debug(f"Taking agcam exposure {self.telescope}-{next_seqno}.")
        cmd = await command.send_command(
            self.agcam,
            f"expose -n {next_seqno} --{flavour} {exposure_time}",
        )

        if cmd.status.did_fail:
            command.actor._status &= ~GuiderStatus.EXPOSING
            command.actor._status |= GuiderStatus.FAILED
            command.actor.status |= GuiderStatus.IDLE

            self.last_seqno = -1

            error = cmd.replies.get("error")

            if nretries > 0:
                command.warning(f"Failed while exposing cameras: {error} Retrying.")
                return await self.expose(
                    command,
                    exposure_time=exposure_time,
                    flavour=flavour,
                    extract_sources=extract_sources,
                    nretries=nretries - 1,
                )

            else:
                raise RuntimeError(f"Failed while exposing cameras: {error}")

        else:
            self.last_seqno = next_seqno

        filenames: set[pathlib.Path] = set()
        for reply in cmd.replies:
            if "filename" in reply.message:
                filename = reply.message["filename"]["filename"]
                cam_name = reply.message["filename"]["camera"]
                filenames.add(pathlib.Path(filename))
                if flavour == "dark":
                    self._write_dark_info(cam_name, filename)

        if len(filenames) == 0:
            raise ValueError("Exposure did not produce any images.")

        if any([self.is_shifted(filename) for filename in filenames]):
            if nretries > 0:
                command.warning("Found shifted images. Retrying.")
                return await self.expose(
                    command,
                    exposure_time=exposure_time,
                    flavour=flavour,
                    extract_sources=extract_sources,
                    nretries=nretries - 1,
                )
            else:
                raise RuntimeError("Run out of retries. Exposing failed.")

        headers: dict[pathlib.Path, fits.Header] = {}

        # Create a new extension with processed data.
        for fn in filenames:
            camname = fits.getval(str(fn), "CAMNAME", "RAW")

            dark_file = self._get_dark_frame(fn, camname)
            if dark_file is None:
                command.warning(f"No dark frame found for camera {camname}.")
                dark_file = ""

            header = header_from_model("PROC")
            header["TELESCOP"] = self.telescope
            header["CAMNAME"] = camname
            header["DARKFILE"] = os.path.basename(dark_file)
            header["DIRNAME"] = str(fn.absolute().parent)
            header["GUIDERV"] = __version__
            header["WCSMODE"] = "none"

            headers[fn] = header

        sources: list[pandas.DataFrame] = []
        if flavour == "object" and extract_sources:
            try:
                sources = await asyncio.gather(
                    *[run_in_executor(extract_sources_func, fn) for fn in filenames]
                )
            except Exception as err:
                command.warning(f"Failed extracting sources: {err}")
                extract_sources = False
            else:
                for ifn, fn in enumerate(filenames):
                    sources_path = fn.with_suffix(".parquet")
                    sources[ifn].to_parquet(sources_path)

                    headers[fn]["SOURCESF"] = sources_path.name

        if len(sources) > 0:
            all_sources = pandas.concat(sources)
            valid = all_sources.loc[all_sources.valid == 1]
            fwhm = numpy.percentile(valid["fwhm"], 25) if len(valid) > 0 else None
        else:
            valid = []
            fwhm = None

        command.info(
            frame={
                "seqno": next_seqno,
                "filenames": [str(fn) for fn in filenames],
                "flavour": flavour,
                "n_sources": len(valid),
                "focus_position": round(focus_position, 1),
                "fwhm": numpy.round(float(fwhm), 3) if fwhm else -999.0,
            }
        )

        command.actor._status &= ~GuiderStatus.FAILED
        command.actor.status &= ~GuiderStatus.EXPOSING

        if not command.actor.status & GuiderStatus.NON_IDLE:
            command.actor.status |= GuiderStatus.IDLE

        with elapsed_time(command, "updating lvm.agcam file"):
            await asyncio.gather(
                *[
                    run_in_executor(update_fits, fn, "PROC", header=headers[fn])
                    for fn in filenames
                ]
            )

        return (list(filenames), next_seqno, list(sources) if extract_sources else None)

    def reset_seqno(self):
        """Resets the seqno.

        This forces a full check of the files next time that
        `.get_next_seqno` is called.

        """

        self.last_seqno = -1

    def get_next_seqno(self):
        """Determines the next exposure sequence number."""

        if get_sjd("LCO") == self.sjd and self.last_seqno >= 0:
            return self.last_seqno + 1

        self.sjd = get_sjd("LCO")

        path = pathlib.Path(f"/data/agcam/{self.sjd}")
        if not path.exists():
            self.last_seqno = 0
            return self.last_seqno + 1

        files = path.glob(f"*{self.telescope}*")
        seqnos = []
        for fn in files:
            se = re.search(r"\_([0-9]+)", str(fn))
            if not se:
                continue
            seqnos.append(int(se.group(1)))

        if len(seqnos) == 0:
            self.last_seqno = 0
        else:
            self.last_seqno = max(seqnos)

        return self.last_seqno + 1

    def is_shifted(self, filename: str | pathlib.Path | numpy.ndarray):
        """Determines if an image is "shifted"."""

        data: numpy.ndarray
        if isinstance(filename, (str, pathlib.Path)):
            data = fits.getdata(str(filename))
        else:
            data = filename

        # If more than 10% of the pixels have values >32,000 this is probably
        # a shifted image.
        npix = data.size
        if (data > 32000).sum() / npix > 0.1:
            return True

        # Another case of a bad image.
        if data.sum() == 0:
            return True

        return False

    async def _get_focus_position(self, command: GuiderCommand):
        """Retrieves the focuser position,"""

        try:
            focus_position = command.actor.models[self.foc]["Position"].value
            assert focus_position is not None
        except Exception:
            command.warning("Focus position not found in model. Querying the actor.")
        else:
            return focus_position

        try:
            focus_cmd = await command.send_command(self.foc, "status", internal=True)
            focus_position = focus_cmd.replies.get("Position")
            assert focus_position is not None
        except Exception as err:
            command.warning(f"Failed getting {self.telescope} focuser position: {err}")
            return -999

        return focus_position

    def _get_dark_frame(self, filename: str | pathlib.Path, cam_name: str):
        """Gets the path to the dark frame."""

        filename = pathlib.Path(filename).absolute()

        dark_file = self.dark_file.get(cam_name, None)
        if dark_file is not None:
            return dark_file

        dirname = filename.parent
        path = dirname / f"lvm.{self.telescope}.agcam.{cam_name}.dark.dat"

        if path.exists():
            dark_file = open(path, "r").read().strip()
            self.dark_file[cam_name] = dark_file
            return pathlib.Path(dark_file)

        return None

    def _write_dark_info(self, cam_name: str, filename: str):
        """Writes a file with the current dark frame."""

        self.dark_file[cam_name] = filename

        dirname = os.path.dirname(filename)
        path = os.path.join(dirname, f"lvm.{self.telescope}.agcam.{cam_name}.dark.dat")
        with open(path, "w") as fd:
            print(str(filename), file=fd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: set_pixel.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from lvmguider.actor import lvmguider_parser


if TYPE_CHECKING:
    from lvmguider.actor import GuiderCommand


__all__ = ["set_pixel", "reset_pixel"]


@lvmguider_parser.command()
@click.argument("PIXEL-X", type=float)
@click.argument("PIXEL-Z", type=float)
async def set_pixel(command: GuiderCommand, pixel_x: float, pixel_z: float):
    """Sets the full frame pixel coordinates on which to guide.

    This command can be issued during active guiding to change the pointing
    of the telescope.

    """

    if not command.actor.guider:
        return command.fail("Guider is not active.")

    command.actor.guider.set_pixel(pixel_x, pixel_z)

    return command.finish(f"Guide pixel is now ({pixel_x:.2f}, {pixel_z:.2f}).")


@lvmguider_parser.command()
async def reset_pixel(command: GuiderCommand):
    """Resets the full frame pixel on which to guide to the central pixel..

    This command can be issued during active guiding to change the pointing
    of the telescope.

    """

    if not command.actor.guider:
        return command.fail("Guider is not active.")

    pixel_x, pixel_z = command.actor.guider.set_pixel()

    return command.finish(f"Guide pixel is now ({pixel_x:.2f}, {pixel_z:.2f}).")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-30
# @Filename: dataclasses.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import warnings
from dataclasses import dataclass, field

from typing import TYPE_CHECKING, Literal

import nptyping as npt
import numpy
import pandas
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS, FITSFixedWarning
from packaging.version import Version

from sdsstools.time import get_sjd

from lvmguider import config, log
from lvmguider.tools import dataframe_from_model, get_frameno
from lvmguider.transformations import get_crota2


if TYPE_CHECKING:
    from pandas._libs.missing import NAType

    from lvmguider.types import ARRAY_2D_F32


__all__ = ["CameraSolution", "GuiderSolution", "FrameData"]


WCS_MODE_T = Literal["none"] | Literal["astrometrynet"] | Literal["gaia"]


@dataclass(kw_only=True)
class BaseSolution:
    """A base class for an object with an astrometric solution."""

    telescope: str
    wcs: WCS | None = None
    matched: bool = False

    def __post_init__(self):
        if self.sources is None:
            self.sources = dataframe_from_model("SOURCES")

    @property
    def solved(self):
        """Was the frame solved?"""

        return self.wcs is not None

    @property
    def pa(self):
        """Returns the position angle from the WCS."""

        if self.wcs is not None:
            return get_crota2(self.wcs)

        return numpy.nan

    @property
    def zero_point(self):
        """Returns the median zero point."""

        if self.sources is not None:
            return float(self.sources.zp.dropna().median())

        return numpy.nan

    @property
    def fwhm(self):
        """Returns the FWHM from the extracted sources."""

        if self.sources is not None:
            fwhms = self.sources.loc[self.sources.valid == 1].fwhm.dropna()
            if len(fwhms) == 0:
                return numpy.nan
            perc_25 = numpy.percentile(fwhms, 25)
            return float(perc_25)

        return numpy.nan

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the camera pointing at its central pixel."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_ag_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])


@dataclass(kw_only=True)
class CameraSolution(BaseSolution):
    """A camera solution, including the determined WCS."""

    frameno: int
    camera: str
    path: pathlib.Path
    date_obs: Time = field(default_factory=lambda: Time.now())
    sources: pandas.DataFrame | None = None
    ref_frame: pathlib.Path | None = None
    wcs_mode: WCS_MODE_T = "none"
    guider_version: Version = Version("0.0.0")

    @classmethod
    def open(cls, file: str | pathlib.Path):
        """Creates an instance from an ``lvm.agcam`` file."""

        file = pathlib.Path(file)
        hdul = fits.open(str(file))

        if "PROC" not in hdul:
            raise ValueError("HDU list does not have a PROC extension.")

        raw = hdul["RAW"].header
        proc = hdul["PROC"].header

        if "GUIDERV" not in proc:
            raise ValueError("Guider version not found.")

        guiderv = Version(proc["GUIDERV"])
        if guiderv < Version("0.4.0a0") and guiderv != Version("0.0.0"):
            raise ValueError(
                "The file was generated with an unsupported version of lvmguider."
            )

        sources: pandas.DataFrame | None = None
        if proc["SOURCESF"]:
            dirname = file.parent
            sources = pandas.read_parquet(dirname / proc["SOURCESF"])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FITSFixedWarning)
            solved = proc["SOLVED"]
            wcs_mode = proc["WCSMODE"].lower()
            wcs = WCS(proc) if solved and wcs_mode != "none" else None

            # Fix cases when WCSMODE is not "none" but the WCS is not valid.
            if wcs is not None and not wcs.is_celestial:
                log.warning(f"Invalid WCS found for {file!s}.")
                wcs = None

        ref_file = proc["REFFILE"]
        ref_frame = pathlib.Path(ref_file) if ref_file is not None else None

        return CameraSolution(
            frameno=get_frameno(file),
            telescope=raw["TELESCOP"],
            camera=raw["CAMNAME"],
            date_obs=Time(raw["DATE-OBS"], format="isot"),
            path=file.absolute(),
            sources=sources,
            wcs=wcs,
            wcs_mode=wcs_mode,
            matched=len(sources.ra.dropna()) > 0 if sources is not None else False,
            ref_frame=ref_frame,
            guider_version=guiderv,
        )


@dataclass(kw_only=True)
class GuiderSolution(BaseSolution):
    """A class to hold an astrometric solution determined by the guider.

    This class abstracts an astrometric solution regardless of whether it was
    determined using astrometry.net or Gaia sources.

    """

    frameno: int
    telescope: str
    solutions: list[CameraSolution]
    guide_pixel: npt.NDArray[npt.Shape["2"], npt.Float32]
    ra_field: float = numpy.nan
    dec_field: float = numpy.nan
    pa_field: float = numpy.nan
    ra_off: float = numpy.nan
    dec_off: float = numpy.nan
    pa_off: float = numpy.nan
    axis0_off: float = numpy.nan
    axis1_off: float = numpy.nan
    correction_applied: bool = False
    correction: list[float] = field(default_factory=lambda: [0.0, 0.0, 0.0])
    guide_mode: str = "guide"
    guider_version: Version = Version("0.0.0")

    @property
    def sources(self):
        """Concatenates sources from the co-added solutions."""

        return pandas.concat(
            [cs.sources for cs in self.solutions if cs.sources is not None],
            axis=0,
        )

    def __getitem__(self, key: str):
        """Returns the associated camera solution."""

        for solution in self.solutions:
            if solution.camera == key:
                return solution

        raise KeyError(f"Invalid camera name {key!r}.")

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the telescope pointing at boresight from the WCS."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_full_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    @property
    def pixel_pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the telescope pointing at the full frame pixel from the WCS."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*self.guide_pixel)
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    @property
    def solved(self):
        """Was the frame solved?"""

        return self.wcs is not None

    @property
    def cameras(self):
        """Returns a list of cameras."""

        return [cs.camera for cs in self.solutions]

    @property
    def separation(self):
        """Returns the separation between field centre and telescope pointing."""

        return numpy.hypot(self.ra_off, self.dec_off)

    @property
    def n_cameras_solved(self):
        """Returns the number of cameras solved."""

        return len([cs for cs in self.solutions if cs.solved])

    def all_cameras_solved(self):
        """Returns `True` if all the cameras have solved."""

        return all([cs.solved for cs in self.solutions])

    @classmethod
    def open(cls, file: str | pathlib.Path, dirname: str | pathlib.Path | None = None):
        """Creates an instance from an ``lvm.guider`` file."""

        file = pathlib.Path(file)
        hdul = fits.open(str(file))

        guider_data = hdul["GUIDERDATA"].header

        if "GUIDERV" not in guider_data:
            raise ValueError("Guider version not found.")

        guiderv = Version(guider_data["GUIDERV"])
        if guiderv < Version("0.4.0a0") and guiderv != Version("0.0.0"):
            raise ValueError(
                "The file was generated with an unsupported version of lvmguider."
            )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FITSFixedWarning)
            wcs = WCS(guider_data) if guider_data["SOLVED"] else None

        solutions: list[CameraSolution] = []
        for key in ["FILEEAST", "FILEWEST"]:
            if guider_data[key] is not None:
                dirname = pathlib.Path(dirname or guider_data["DIRNAME"] or file.parent)
                solutions.append(CameraSolution.open(dirname / guider_data[key]))

        if "XFFPIX" in guider_data:
            guide_pixel = numpy.array([guider_data["XFFPIX"], guider_data["ZFFPIX"]])
        elif "XMFPIX" in guider_data:
            # From when we called it master frame.
            guide_pixel = numpy.array([guider_data["XMFPIX"], guider_data["ZMFPIX"]])
        else:
            log.warning("Missing pixel information. Default to central pixel.")
            guide_pixel = config["xz_full_frame"]

        return GuiderSolution(
            frameno=get_frameno(file),
            telescope=guider_data["TELESCOP"],
            solutions=solutions,
            guide_pixel=guide_pixel,
            wcs=wcs,
            ra_off=guider_data["OFFRAMEA"] or numpy.nan,
            dec_off=guider_data["OFFDEMEA"] or numpy.nan,
            pa_off=guider_data["OFFPAMEA"] or numpy.nan,
            axis0_off=guider_data["OFFA0MEA"] or numpy.nan,
            axis1_off=guider_data["OFFA1MEA"] or numpy.nan,
            correction_applied=guider_data["CORRAPPL"],
            correction=[
                guider_data["AX0CORR"],
                guider_data["AX1CORR"],
                guider_data["PACORR"],
            ],
            guider_version=guiderv,
        )


@dataclass(kw_only=True)
class FrameData:
    """Data associated with a frame."""

    frameno: int
    path: pathlib.Path
    raw_header: dict
    date_obs: Time
    sjd: int
    camera: str
    telescope: str
    exptime: float
    image_type: str
    kmirror_drot: float
    focusdt: float
    solution: CameraSolution
    stacked: bool = False
    reprocessed: bool = False
    data: ARRAY_2D_F32 | None = None

    @property
    def sources(self):
        """Accessor to the camera solution sources."""

        return self.solution.sources


class CoAddWarningsMixIn:
    """Methods to calculate warnings for co-added data classes."""

    zero_point: float
    fwhm: float
    sources: pandas.DataFrame

    def transp_warning(self):
        """Determines whether to raise a transparency warning."""

        # We raise a transparency warning when the median ZP is > some nominal plus
        # a number of magnitudes. The nominal ZP corresponds to a photometric night.
        nominal_zp = config["coadds"]["warnings"]["nominal_zp"]
        zp_overmag_warning = config["coadds"]["warnings"]["zp_overmag_warning"]

        if numpy.isnan(self.zero_point):
            return True
        else:
            return self.zero_point > (nominal_zp + zp_overmag_warning)

    def fwhm_warning(self):
        """Determines whether to raise a transparency warning."""

        # Similarly, for FWHM we raise a warning if the co-added FWHM
        # is > factor * median FWHM. This may indicate that there was a jump
        # during guiding.
        if numpy.isnan(self.zero_point):
            return True
        else:
            fwhm_factor_warn = config["coadds"]["warnings"]["fwhm_factor_warning"]
            sources_fwhm = self.sources.loc[self.sources.valid == 1, "fwhm"].dropna()
            if len(sources_fwhm) == 0:
                return False
            perc_25 = numpy.percentile(sources_fwhm, 25)
            return self.fwhm > perc_25 * fwhm_factor_warn


@dataclass(kw_only=True)
class CoAdd_CameraSolution(CameraSolution, CoAddWarningsMixIn):
    """A camera solution for a co-added frame."""

    path: pathlib.Path = pathlib.Path("")
    coadd_image: ARRAY_2D_F32 | None = None
    frames: list[FrameData] | None = None
    sources: pandas.DataFrame | None = None
    sigmaclip: bool = False
    sigmaclip_sigma: float | None = None

    def frame_data(self):
        """Returns a Pandas data frame from a list of `.FrameData`."""

        df = dataframe_from_model("FRAMEDATA")

        if self.frames is None:
            return df

        for ii, fd in enumerate(self.frames):
            new_row = dict(
                frameno=fd.frameno,
                mjd=fd.sjd,
                date_obs=fd.date_obs.isot,
                camera=fd.camera,
                telescope=fd.telescope,
                exptime=numpy.float32(fd.exptime),
                kmirror_drot=numpy.float32(fd.kmirror_drot),
                focusdt=numpy.float32(fd.focusdt),
                fwhm=numpy.float32(fd.solution.fwhm),
                pa=numpy.float32(fd.solution.pa),
                zero_point=numpy.float32(fd.solution.zero_point),
                stacked=int(fd.stacked),
                solved=int(fd.solution.solved),
                wcs_mode=fd.solution.wcs_mode,
            )
            df.loc[ii, list(new_row)] = list(new_row.values())

        df = df.sort_values(["frameno", "camera"])

        df.reset_index(drop=True, inplace=True)

        return df


@dataclass(kw_only=True)
class GlobalSolution(BaseSolution, CoAddWarningsMixIn):
    """A global solution from co-added frames."""

    coadd_solutions: list[CoAdd_CameraSolution]
    guider_solutions: list[GuiderSolution]
    path: pathlib.Path = pathlib.Path("")
    matched: bool = True

    @property
    def n_cameras_solved(self):
        """Returns the number of cameras solved."""

        return len([cs for cs in self.coadd_solutions if cs.solved])

    @property
    def sources(self):
        """Concatenates sources from the co-added solutions."""

        return pandas.concat(
            [cs.sources for cs in self.coadd_solutions if cs.sources is not None],
            axis=0,
        )

    @property
    def pointing(self) -> npt.NDArray[npt.Shape["2"], npt.Float64]:
        """Returns the camera pointing at its central pixel."""

        if not self.wcs:
            return numpy.array([numpy.nan, numpy.nan])

        skyc = self.wcs.pixel_to_world(*config["xz_full_frame"])
        return numpy.array([skyc.ra.deg, skyc.dec.deg])

    def frame_data(self):
        """Concatenates the frame data and returns a Pandas data frame."""

        frame_data = pandas.concat([cs.frame_data() for cs in self.coadd_solutions])
        frame_data.sort_values(["frameno", "camera"], inplace=True)

        return frame_data

    def guider_data(self):
        """Constructs a data frame with guider data."""

        df = dataframe_from_model("GUIDERDATA_FRAME")

        for ii, gs in enumerate(self.guider_solutions):
            pa = get_crota2(gs.wcs) if gs.wcs is not None else numpy.nan
            pointing = gs.pointing

            date_obs: str = ""
            mjd: int | NAType = pandas.NA

            if len(gs.solutions) > 0:
                date_obs = gs.solutions[0].date_obs.isot
                mjd = get_sjd("LCO", gs.solutions[0].date_obs.to_datetime())

            new_row = dict(
                frameno=gs.frameno,
                telescope=gs.telescope,
                mjd=mjd,
                date_obs=date_obs,
                fwhm=numpy.float32(gs.fwhm),
                pa=numpy.float32(pa),
                zero_point=numpy.float32(gs.zero_point),
                solved=int(gs.solved),
                n_cameras_solved=int(gs.n_cameras_solved),
                guide_mode=gs.guide_mode,
                x_ff_pixel=numpy.float32(gs.guide_pixel[0]),
                z_ff_pixel=numpy.float32(gs.guide_pixel[1]),
                ra=numpy.float64(pointing[0]),
                dec=numpy.float64(pointing[1]),
                ra_field=numpy.float64(gs.ra_field),
                dec_field=numpy.float64(gs.dec_field),
                pa_field=numpy.float32(gs.pa_field),
                ra_off=numpy.float32(gs.ra_off),
                dec_off=numpy.float32(gs.dec_off),
                pa_off=numpy.float32(gs.pa_off),
                axis0_off=numpy.float32(gs.axis0_off),
                axis1_off=numpy.float32(gs.axis1_off),
                applied=int(gs.correction_applied),
                ax0_applied=numpy.float32(gs.correction[0]),
                ax1_applied=numpy.float32(gs.correction[1]),
                rot_applied=numpy.float32(gs.correction[2]),
            )
            df.loc[ii, list(new_row)] = list(new_row.values())

        df = df.sort_values(["frameno"])
        df.reset_index(drop=True, inplace=True)

        return df
